import attr


@attr.s()
class Issue:
    id: int = attr.ib()
    title: str = attr.ib()
    description: str = attr.ib()
    web_url: str = attr.ib(default=None)
    reference: str = attr.ib(default=None)



import asyncio
from typing import Annotated

import typer
from rich.console import Console
from rich.text import Text

from issx.clients import SupportedBackend
from issx.clients.gitlab import GitlabClient, GitlabInstanceClient
from issx.clients.redmine import RedmineClient, RedmineInstanceClient
from issx.instance_managers import GenericConfigParser, InstanceManager
from issx.services import CopyIssueService

app = typer.Typer(no_args_is_help=True)

BackendOption = Annotated[SupportedBackend, typer.Option()]
InstanceNameOption = Annotated[str, typer.Option("--instance")]
ProjectOption = Annotated[
    str, typer.Option("--project", help="Project name configured in the config file")
]
console = Console()

InstanceManager.register_backend(
    SupportedBackend.gitlab, GitlabInstanceClient, GitlabClient
)
InstanceManager.register_backend(
    SupportedBackend.redmine, RedmineInstanceClient, RedmineClient
)


@app.command()
def copy(
    source_project_name: Annotated[
        str,
        typer.Option(
            "--source", help="Source project name configured in the config file"
        ),
    ],
    target_project_name: Annotated[
        str,
        typer.Option(
            "--target", help="Target project name configured in the config file"
        ),
    ],
    issue_id: int,
    title_format: Annotated[
        str,
        typer.Option(
            "--title-format",
            "-T",
            help="Template of a new issue title. Can contain placeholders of the"
            " issue attributes: {id}, {title}, {description}, {web_url}, {reference}",
        ),
    ] = "{title}",
    description_format: Annotated[
        str,
        typer.Option(
            "--description-format",
            "-D",
            help="Template of a new issue description. Can contain placeholders of the"
            " issue attributes: {id}, {title}, {description}, {web_url}, {reference}",
        ),
    ] = "{description}",
    allow_duplicates: Annotated[
        bool,
        typer.Option(
            "--allow-duplicates",
            "-A",
            help="Allow for duplicate issues. If set, the command will return the first"
            " issue found with the same title. If no issues are found,"
            " a new issue will be created.",
        ),
    ] = False,
    assign_to_me: Annotated[bool, typer.Option("--assign-to-me", "-M")] = False,
) -> int:
    console.print(
        Text.assemble(
            f"Copying issue {issue_id} from project ",
            (source_project_name, "bold magenta"),
            " to project ",
            (target_project_name, "bold magenta"),
        )
    )
    config = GenericConfigParser()
    instance_manager = InstanceManager(config)
    try:
        source_client = instance_manager.get_project_client(source_project_name)
        target_client = instance_manager.get_project_client(target_project_name)
    except Exception as e:
        console.print_exception()
        console.print("Error when configuring client instance.\n", style="red")
        raise typer.Exit(1) from e
    new_issue = asyncio.run(
        CopyIssueService(source_client, target_client).copy(
            issue_id,
            title_format,
            description_format,
            allow_duplicates=allow_duplicates,
            assign_to_me=assign_to_me,
        )
    )
    console.print(f"Success!\n{new_issue}", style="green")
    return 0


@app.command()
def auth_verify(instance_name: InstanceNameOption) -> None:
    config = GenericConfigParser()
    instance_manager = InstanceManager(config)
    try:
        instance = instance_manager.get_instance_client(instance_name)
    except Exception as e:
        console.print_exception()
        console.print("Error when configuring client instance.\n", style="red")
        raise typer.Exit(1) from e
    try:
        username = asyncio.run(instance.auth())
        if not username:
            raise Exception("Authentication failed")
    except Exception as e:
        console.print_exception()
        console.print(
            f"Error when authenticating to {instance.get_instance_url()}",
            style="red",
        )
        raise typer.Exit(1) from e
    else:
        console.print(
            "Authentication successful",
            f"Instance: {instance.get_instance_url()}",
            f"User: {username}",
            sep="\n",
            style="green bold italic",
        )


if __name__ == "__main__":
    app()



from typing import Self, cast

from gitlab import Gitlab, GitlabGetError
from gitlab.v4.objects import CurrentUser, Project, ProjectIssue

from issx.clients.exceptions import IssueDoesNotExistError, ProjectDoesNotExistError
from issx.clients.interfaces import InstanceClientInterface, IssueClientInterface
from issx.domain.issues import Issue


class IssueMapper:
    """
    Maps Gitlab API objects to domain objects
    """

    @classmethod
    def issue_to_domain(cls, issue: ProjectIssue) -> Issue:
        return Issue(
            id=issue.iid,
            title=issue.title,
            description=issue.description,
            web_url=issue.web_url,
            reference=issue.references["full"],
        )

    @classmethod
    def issues_to_domain_list(cls, issues: list[ProjectIssue]) -> list[Issue]:
        return [cls.issue_to_domain(issue) for issue in issues]


class GitlabInstanceClient(InstanceClientInterface):
    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls(Gitlab(config["url"], private_token=config["token"]))

    def __init__(self, client: Gitlab):
        self.client = client

    async def auth(self) -> str | None:
        self.client.auth()
        if user := self.client.user:
            return cast(str, user.username)
        return None

    def get_user(self) -> CurrentUser:
        self.client.auth()
        if not self.client.user:
            raise ValueError("Cannot get user from Gitlab client")
        return self.client.user

    def get_instance_url(self) -> str:
        return self.client.url


class GitlabClient(IssueClientInterface, GitlabInstanceClient):
    def __init__(self, client: Gitlab, project_id: int):
        self.project_id = project_id
        self._project: Project | None = None
        super().__init__(client)

    async def create_issue(
        self, title: str, description: str, assign_to_me: bool = False
    ) -> Issue:
        project = await self._get_project()
        issue: ProjectIssue = cast(
            ProjectIssue,
            project.issues.create(
                {
                    "title": title,
                    "description": description,
                    "assignee_id": self.get_user().id,
                }
            ),
        )
        return IssueMapper.issue_to_domain(issue)

    async def get_issue(self, issue_id: int) -> Issue:
        issue: ProjectIssue = await self._get_issue(issue_id)
        return IssueMapper.issue_to_domain(issue)

    async def find_issues(self, title: str) -> list[Issue]:
        project = await self._get_project()
        issues = cast(list[ProjectIssue], list(project.issues.list(search=title)))
        return IssueMapper.issues_to_domain_list(issues)

    async def _get_project(self) -> Project:
        if self._project is None:
            try:
                self._project = self.client.projects.get(self.project_id)
            except GitlabGetError as e:
                raise ProjectDoesNotExistError(
                    f"Project with id={self.project_id} does not exist"
                ) from e
        return self._project

    async def _get_issue(self, issue_id: int) -> ProjectIssue:
        project = await self._get_project()
        try:
            return project.issues.get(issue_id)
        except GitlabGetError as e:
            raise IssueDoesNotExistError(issue_id) from e

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls(
            GitlabInstanceClient.from_config(config["instance"]).client,
            project_id=int(config["project"]),
        )



from typing import Self

from redminelib import Redmine
from redminelib.exceptions import ResourceNotFoundError
from redminelib.resources import Issue as RedmineIssue
from redminelib.resources import Project
from redminelib.resultsets import ResourceSet

from issx.clients.exceptions import IssueDoesNotExistError, ProjectDoesNotExistError
from issx.clients.interfaces import InstanceClientInterface, IssueClientInterface
from issx.domain.issues import Issue


class RedmineIssueMapper:
    """
    Maps Redmine API objects to domain objects
    """

    @classmethod
    def issue_to_domain(cls, issue: RedmineIssue) -> Issue:  # type: ignore[no-any-unimported]
        return Issue(
            id=issue.id,
            title=issue.subject,
            description=issue.description,
            web_url=issue.url,
            reference=issue.id,
        )

    @classmethod
    def issues_to_domain_list(cls, issues: ResourceSet) -> list[Issue]:  # type: ignore[no-any-unimported]
        return [cls.issue_to_domain(issue) for issue in issues]


class RedmineInstanceClient(InstanceClientInterface):
    def __init__(self, client: Redmine):  # type: ignore[no-any-unimported]
        self.client = client

    async def auth(self) -> str | None:
        return str(self.client.auth())

    def get_instance_url(self) -> str:
        return str(self.client.url)

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls(
            Redmine(
                config["url"],
                key=config["token"],
            )
        )


class RedmineClient(IssueClientInterface, RedmineInstanceClient):
    def __init__(self, client: Redmine, project_id: int):  # type: ignore[no-any-unimported]
        self._project_id = project_id
        self._project: Project | None = None  # type: ignore[no-any-unimported]
        super().__init__(client)

    async def create_issue(
        self, title: str, description: str, assign_to_me: bool = False
    ) -> Issue:
        issue = self.client.issue.create(
            project_id=(await self.get_project()).id,
            subject=title,
            description=description,
            assigned_to_id="me" if assign_to_me else None,
        )
        return RedmineIssueMapper.issue_to_domain(issue)

    async def get_issue(self, issue_id: int) -> Issue:
        try:
            issue = self.client.issue.get(issue_id)
        except ResourceNotFoundError as e:
            raise IssueDoesNotExistError(issue_id) from e
        return RedmineIssueMapper.issue_to_domain(issue)

    async def find_issues(self, title: str) -> list[Issue]:
        return RedmineIssueMapper.issues_to_domain_list(
            self.client.issue.filter(
                project_id=(await self.get_project()).id, subject=title
            )
        )

    async def get_project(self) -> Project:  # type: ignore[no-any-unimported]
        if self._project is None:
            try:
                self._project = self.client.project.get(self._project_id)
            except ResourceNotFoundError as e:
                raise ProjectDoesNotExistError(
                    f"Project with id={self._project_id} does not exist"
                ) from e
        return self._project

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls(
            RedmineInstanceClient.from_config(config["instance"]).client,
            project_id=int(config["project"]),
        )



import abc
from typing import Self

from issx.domain.issues import Issue


class InstanceClientInterface(abc.ABC):
    """
    Interface for a client that interacts with an issue tracker instance
    """

    @abc.abstractmethod
    async def auth(self) -> str | None:
        """
        Authenticate the client with the instance.
        Raises an internal error if the authentication fails or returns None.
        :return: The username of the authenticated user or None
        """
        pass

    @abc.abstractmethod
    def get_instance_url(self) -> str:
        """
        :return: The URL of the instance
        """
        pass

    @classmethod
    @abc.abstractmethod
    def from_config(cls, config: dict) -> Self:
        """
        Create an instance of the client from a configuration dictionary.
        :param config: The configuration dictionary. Higher level code should validate
        the configuration.
        :return: An instance of the client
        """
        pass


class IssueClientInterface(abc.ABC):
    @abc.abstractmethod
    async def get_issue(self, issue_id: int) -> Issue:
        """
        Retrieve an issue by its ID.
        Raises IssueDoesNotExistError if the issue does not exist.
        :param issue_id: The ID of the issue
        :return: Issue object
        """
        pass

    @abc.abstractmethod
    async def create_issue(
        self, title: str, description: str, assign_to_me: bool = False
    ) -> Issue:
        """
        Create a new issue.
        :param title: The title of the issue
        :param description: The description of the issue
        :param assign_to_me: Assign the issue to the authenticated user
        :return:
        """
        pass

    @abc.abstractmethod
    async def find_issues(self, title: str) -> list[Issue]:
        """
        Find issues by title using an exact match.
        :param title: The title of the issue
        :return: List of issues
        """
        pass

    @classmethod
    @abc.abstractmethod
    def from_config(cls, config: dict) -> Self:
        pass



import tomllib
from pathlib import Path
from typing import Any

from issx.clients import SupportedBackend
from issx.clients.interfaces import InstanceClientInterface, IssueClientInterface


class GenericConfigParser:
    def __init__(self, config_file: Path | None = None):
        self.config_file = config_file or self._find_config_file()
        self._data: dict[str, dict[str, dict]] = tomllib.loads(
            self.config_file.read_text()
        )

    def get_instance_config(self, instance: str) -> dict:
        return self._data["instances"][instance]

    def get_project_config(self, project: str) -> dict:
        return self._data["projects"][project]

    def __getitem__(self, item: str) -> Any:
        return self._data[item]

    @staticmethod
    def _find_config_file() -> Path:
        locations = [
            Path("issx.toml"),
            Path("~/.config/issx.toml").expanduser(),
        ]
        for location in locations:
            if location.exists() and location.is_file() and location.suffix == ".toml":
                return location
        raise FileNotFoundError("No configuration file found")


class InstanceManager:
    backends: dict[
        SupportedBackend,
        tuple[type[InstanceClientInterface], type[IssueClientInterface]],
    ] = {}

    def __init__(self, config: GenericConfigParser):
        self.config = config

    @classmethod
    def register_backend(
        cls,
        backend: SupportedBackend,
        instance_client_class: type[InstanceClientInterface],
        project_client_class: type[IssueClientInterface],
    ) -> None:
        cls.backends[backend] = (instance_client_class, project_client_class)

    def get_instance_client(self, instance: str) -> InstanceClientInterface:
        instance_config = self.config.get_instance_config(instance)
        backend = SupportedBackend(instance_config["backend"])
        client_class = self.backends[backend][0]
        return client_class.from_config(instance_config)

    def get_project_client(self, project: str) -> IssueClientInterface:
        project_config = self.config.get_project_config(project)
        instance_config = self.config.get_instance_config(project_config["instance"])
        project_config["instance"] = instance_config
        backend = SupportedBackend(instance_config["backend"])
        client_class = self.backends[backend][1]
        return client_class.from_config(project_config)



from enum import StrEnum

__all__ = ["SupportedBackend", "GitlabClient"]

from issx.clients.gitlab import GitlabClient


class SupportedBackend(StrEnum):
    gitlab = "gitlab"
    redmine = "redmine"
